﻿using System;
using System.Collections.Generic;
using System.Dynamic;
using System.IO;
using System.Linq;
using System.Text;
using System.Threading.Tasks;

namespace Merthsoft.DynamicConfig {
	/// <summary>
	/// The exception that is thrown when an INI file is invalid.
	/// </summary>
	[Serializable]
	public class IniException : Exception {
		/// <summary>
		/// Gets the line number where the error occurred.
		/// </summary>
		public int LineNumber { get; protected set; }

		/// <summary>
		/// Initializes a new instance of the LineNumber class with a specified line number.
		/// </summary>
		/// <param name="lineNumber">The line number where the error occurred.</param>
		public IniException(int lineNumber) { LineNumber = lineNumber; }

		/// <summary>
		/// Initializes a new instance of the LineNumber class with a specified line number and error message.
		/// </summary>
		/// <param name="message">The error description.</param>
		/// <param name="lineNumber">The line number where the error occurred.</param>
		public IniException(string message, int lineNumber) : base(message) { LineNumber = lineNumber; }

		/// <summary>
		/// Initializes a new instance of the LineNumber class with a specified line number, inner exception, and error message.
		/// </summary>
		/// <param name="message">The error description.</param>
		/// <param name="inner">The Exception that threw the XmlException, if any. This value can be null.</param>
		/// <param name="lineNumber">The line number where the error occurred.</param>
		public IniException(string message, Exception inner, int lineNumber) : base(message, inner) { LineNumber = lineNumber; }

		/// <summary>
		/// Initializes a new instance of the LineNumber class using the information in the SerializationInfo and StreamingContext objects.
		/// </summary>
		/// <param name="info">The SerializationInfo object containing all the properties of an XmlException. </param>
		/// <param name="context">The StreamingContext object containing the context information. </param>
		protected IniException(
		  System.Runtime.Serialization.SerializationInfo info,
		  System.Runtime.Serialization.StreamingContext context)
			: base(info, context) { }
	}

	/// <summary>
	/// Contains the static config methods.
	/// </summary>
	public static class Config {
		/// <summary>
		/// Reads an INI file from a stream.
		/// </summary>
		/// <param name="reader">The stream to read from.</param>
		/// <returns>A dynamic object with properties defined by the INI file.</returns>
		public static dynamic ReadIni(StreamReader reader) {
			bool lineIsSection = false;
			string sectionName;
			int lineNumber = -1;

			IDictionary<string, object> config = new ExpandoObject();
			IDictionary<string, object> currentSection = null;

			while (!reader.EndOfStream) {
				string line = reader.ReadLine().TrimStart();
				lineNumber++;
				if (string.IsNullOrWhiteSpace(line) || line[0] == ';' || line[0] == '#') { continue; }

				lineIsSection = line[0] == '[' && line[line.Length - 1] == ']';

				if (lineIsSection) {
					sectionName = line.Substring(1, line.Length - 2);
					currentSection = new ExpandoObject();
					config.Add(sectionName, currentSection);
				} else {
					if (currentSection == null) { throw new IniException("Current section is null, but trying to read a key.", lineNumber); }

					int equalsLocation = line.IndexOf('=');
					if (equalsLocation == -1) { equalsLocation = line.IndexOf(':'); }
					if (equalsLocation <= 0 || equalsLocation == line.Length - 1) { throw new IniException("Key-value pair not in format 'key=value' or 'key:value'.", lineNumber); }

					string key = line.Substring(0, equalsLocation).Trim();
					string value = line.Substring(equalsLocation + 1).Trim();
					if (value[0] == '"' && value[value.Length - 1] == '"') { value = value.Substring(1, value.Length - 2); }

					currentSection.Add(key, value);
				}
			}

			return (dynamic)config;
		}

		/// <summary>
		/// Reads an INI file from disk.
		/// </summary>
		/// <param name="path">The location of the INI file.</param>
		/// <returns>A dynamic object with properties defined by the INI file.</returns>
		public static dynamic ReadIni(string path) {
			using (StreamReader sr = new StreamReader(path)) {
				return ReadIni(sr);
			}
		}

		/// <summary>
		/// Writes an INI file to disk.
		/// </summary>
		/// <param name="config">The config data to write.</param>
		/// <param name="path">The path to write to.</param>
		public static void WriteIni(this IDictionary<string, object> config, string path) {
			using (StreamWriter sw = new StreamWriter(path)) {
				WriteIni(config, sw);
			}
		}

		/// <summary>
		/// Writes an INI file to a stream.
		/// </summary>
		/// <param name="config">The config data to write.</param>
		/// <param name="sw">The stream to write to.</param>
		private static void WriteIni(this IDictionary<string, object> config, StreamWriter sw) {
			foreach (var prop in config) {
				sw.WriteLine("[{0}]", prop.Key);
				foreach (var item in (IDictionary<string, object>)prop.Value) {
					sw.WriteLine("{0}={1}", item.Key, item.Value);
				}
				sw.WriteLine();
			}
		}

		/// <summary>
		/// Writes an INI file to disk.
		/// </summary>
		/// <param name="config">The config data to write.</param>
		/// <param name="path">The path to write to.</param>
		public static void WriteIni(dynamic config, string path) {
			((IDictionary<string, object>)config).WriteIni(path);
		}

		/// <summary>
		/// Writes an INI file to a stream.
		/// </summary>
		/// <param name="config">The config data to write.</param>
		/// <param name="sw">The stream to write to.</param>
		private static void WriteIni(dynamic config, StreamWriter sw) {
			((IDictionary<string, object>)config).WriteIni(sw);
		}

		/// <summary>
		/// Reloads an INI file into an object.
		/// </summary>
		/// <param name="config">The existing config object to load.</param>
		/// <param name="reader">The reader stream to read from.</param>
		public static void ReloadIni(this IDictionary<string, object> config, StreamReader reader) {
			IDictionary<string, object> newConfig = ReadIni(reader);

			pruneDictionary(config, newConfig);

			foreach (var prop in newConfig) {
				if (config.ContainsKey(prop.Key)) {
					pruneDictionary((IDictionary<string, object>)config[prop.Key], (IDictionary<string, object>)prop.Value);
					foreach (var item in (IDictionary<string, object>)prop.Value) {
						((IDictionary<string, object>)config[prop.Key])[item.Key] = item.Value;
					}
				} else {
					config.Add(prop);
				}
			}
		}

		/// <summary>
		/// Reloads an INI file into an object.
		/// </summary>
		/// <param name="config">The existing config object to load.</param>
		/// <param name="path">The location of the INI file.</param>
		public static void ReloadIni(this IDictionary<string, object> config, string path) {
			using (StreamReader sr = new StreamReader(path)) {
				ReloadIni(config, sr);
			}
		}

		/// <summary>
		/// Reloads an INI file into an object.
		/// </summary>
		/// <param name="config">The existing config object to load.</param>
		/// <param name="path">The location of the INI file.</param>
		public static void ReloadIni(dynamic config, string path) {
			((IDictionary<string, object>)config).ReloadIni(path);
		}
		
		/// <summary>
		/// Reloads an INI file into an object.
		/// </summary>
		/// <param name="config">The existing config object to load.</param>
		/// <param name="reader">The reader stream to read from.</param>
		public static void ReloadIni(dynamic config, StreamReader reader) {
			((IDictionary<string, object>)config).ReloadIni(reader);
		}

		/// <summary>
		/// Prunes non-existing values out of an existing dictionary.
		/// </summary>
		/// <param name="config">The dictionary to prune.</param>
		/// <param name="newConfig">The dictionary to match values from.</param>
		private static void pruneDictionary(IDictionary<string, object> config, IDictionary<string, object> newConfig) {
			ICollection<string> configKeys = config.Keys;
			int configKeysCount = configKeys.Count;
			for (int i = 0; i < configKeysCount; i++) {
				string key = configKeys.ElementAt(i);
				if (!newConfig.ContainsKey(key)) {
					config.Remove(key);
				}
			}
		}
	}
}
